# -*- perl -*-
#
# Copyright (C) 1996 Ken MacLeod
# See the file COPYING for distribution terms.
#

use SGMLS;
use Text::GDO::DocObject;
use Text::GDO::SpecTransform;

# forward declare these
package Text::GDO::SGML::Entity;
package Text::GDO::SGML::Event;
package Text::GDO::SGML::Element;
package Text::GDO::SGML::Notation;

package Text::GDO::SGML;
use File::Basename;

#
# `new' will run `nsgmls' on the `file' or `source' scalar and to load
# and return the instance
#
sub new {
    my ($type, %init) = @_;
    my ($self);

    $self = \%init;

    bless $self, 'Text::GDO::SGML';

    # XXX use a cache
    # XXX use an entity manager

    if (defined $self->{'file'}) {
	my ($file_name) = $self->{'file'};

	# try `.sgml' and `.sgm' if `$file_name' isn't found
	if (!-r $file_name) {
	    if (-f "$file_name.sgml") {
		$file_name .= ".sgml";
	    } elsif (-f "$file_name.sgm") {
		$file_name .= ".sgm";
	    } else {
		warn "$::prog: $file_name: file not found\n";
		return undef;
	    }
	}

	if (!open (ESIS, "$Text::GDO::NSGMLS $Text::GDO::NSGMLSFLAGS $file_name |")) {
	    warn "$::prog: can't run $Text::GDO::NSGMLS on \`$file_name'\n";
	    return undef;
	}

	($self->{'file_name'}, $self->{'file_dir'}) = fileparse ($file_name);

	# XXX $::use_SGMLSpm
	my ($retval) = $self->parse_nsgmls (ESIS, $file_name);
	close (ESIS);

	$self->identify ($file_name);

	if (!$retval) {
	    return undef;
	}
    } elsif (defined $self->{'source'}) {
	my ($retval);

	my ($pid) = open (ESIS, "-|");
	if ($pid == 0) {
	    # XXX $::sgmls
	    if (!open (SGML, "| $Text::GDO::NSGMLS $Text::GDO::NSGMLSFLAGS")) {
		warn "$::prog: can't run nsgmls on internal source\n";
		exit 1;
	    }
	    print SGML $self->{'source'};
	    close (SGML);
	    exit 0;
	} else {
	    # XXX $::use_SGMLSpm
	    $retval = $self->parse_nsgmls (ESIS, '-');
	}
	close (ESIS);

	$self->identify_scalar ($self->{'source'});

	if (!$retval) {
	    return undef;
	}
    }

    return $self;
}

#
# Parse the `ESIS' information coming from `nsgmls'
#
# `parse_nsgmls' is a misnomer.  `parse_nsgmls' is a ``hardcoded'',
# ``unrolled'', version of `parse_sgmls' that doesn't use SGMLSpm.
# It's about five times as fast as using SGMLSpm but requires more
# work later.  XXX Eventually the whole shooting match should be
# replaced by a direct-to-SP library.
#
sub parse_nsgmls {
    my ($self, $file, $file_name) = @_;
    my (@attributes, $pubid, $sysid, @files);

    $::no_SGMLSpm = 1; # allow for ugly hacks later
    my (@instance_stack);
    my $line = 0;
    my $instance = ['start_element', undef, undef, undef, []];
    bless $instance, 'Text::GDO::SGML::Event';
    $self->{content} = $instance;
    while (<$file>) {
	chop;
	my ($op) = substr ($_, 0, 1);
	($op eq 'A') && do {	# Attribute
	    substr ($_, 0, 1) = "";
	    push (@attributes, $_);

	    next;
	};

	($op eq '(') && do {	# Start element
	    my ($element) = new Text::GDO::SGML::Element (substr ($_, 1), $instance, @attributes);
	    my ($event) = new_arr Text::GDO::SGML::Event ('start_element', $element, undef, $line, []);
	    push (@{$instance->[4]}, $event);
	    push (@instance_stack, $instance);
	    $instance = $event;
	    @attributes = ();

	    next;
	};

	($op eq ')') && do {	# End element
	    $instance = pop (@instance_stack);

	    next;
	};

	($op eq 'L') && do {	# Line number
	    $line = substr($_, 1);

	    next;
	};

	($op eq '-') && do { 	# Data
	    my ($event) = new_arr Text::GDO::SGML::Event ('cdata', substr ($_, 1), undef, $line);
	    push (@{$instance->[4]}, $event);

	    next;
	};

	($op eq 's') && do {	# sysid
	    $sysid = substr ($_, 1);

	    next;
	};

	($op eq 'p') && do {	# pubid
	    $pubid = substr ($_, 1);

	    next;
	};

	($op eq 'f') && do {	# file
	    push (@files, substr ($_, 1));

	    next;
	};

	($op eq 'E') && do { 	# External entity
	    my ($entity_data) = substr ($_, 1);
	    $entity_data =~ /^(\S+) (\S+) (\S+)$/
		|| die "$::prog: bad external entity event data: $entity_data\n";
	    my ($name,$type,$notation) = ($1,$2,$3);
	    $self->{'entities'}{$name} = new Text::GDO::SGML::Entity ($name, $type, '', $sysid, $pubid, [@files], $notation);

	    undef $sysid; undef $pubid; undef @files;
	    next;
	};

	($op eq '&') && do {	# external entity reference
	    my ($entity_name) = substr ($_, 1);
	    my ($event) = new_arr Text::GDO::SGML::Event('entity',
							 ($self->{'entities'}->{$entity_name}
							  || die "$::prog: Unknown external entity: $entity_name\n"),
							 undef, $line);
	    push (@{$instance->[4]}, $event);

	    next;
	};
      
	($op eq 'N') && do {	# Notation
	    my ($name) = substr ($_, 1);
	    $self->{'notations'}{$name}
	        = new Text::GDO::SGML::Notation ($name, $sysid, $pubid);

	    undef $sysid; undef $pubid; undef @files;
	    next;
	};

	($op eq 'S') && do {	# Subdoc name
	    warn "$::prog: ESIS command character \`$op' not implemented\n";

	    undef $sysid; undef $pubid; undef @files;
	    next;
	};

	($op eq 'T') && do {	# Subdoc name
	    warn "$::prog: ESIS command character \`$op' not implemented\n";

	    undef $sysid; undef $pubid; undef @files;
	    next;
	};

	($op eq 'C') && do {	# Conforming
	    next;
	};

	warn "$::prog: ESIS command character \`$op' not recognized\n";
    }

    if ($#{$instance->[4]} < 0) {
	warn "$::prog: empty input from \`$Text::GDO::NSGMLS $file_name'\n";
	return 0;
    } else {
	return 1;
    }
}

#
# Parse the `ESIS' information coming from `nsgmls' using `SGMLSpm'
#
sub parse_sgmls {
    my ($self, $file, $file_name) = @_;
    my ($parse, $event);

    $parse = new SGMLS($file);

    my $instance = ['start_element', undef, undef, undef, []];
    my (@instance_stack);
    bless $instance, 'Text::GDO::SGML::Event';
    $self->{content} = $instance;
    while ($event = $parse->next_event) {
	my ($event_info) = new Text::GDO::SGML::Event($event);
      SWITCH: {

	  ($event->type eq 'start_element') && do {
	      my $element = $event->data;    # An object of class SGMLS_Element

	      push (@{$instance->[4]}, $event_info);
	      push (@instance_stack, $instance);
	      $event_info->[4] = [];
	      $instance = $event_info;

	      last SWITCH;
	  };

	  ($event->type eq 'end_element') && do {
	      my $element = $event->data;    # An object of class SGMLS_Element

	      $instance = pop (@instance_stack);

	      last SWITCH;
	  };

	  ($event->type eq 'start_subdoc') && do {
	      my $entity = $event->data;     # An object of class SGMLS_Entity
	      #[[your code for the beginning of a subdoc entity]]
	      last SWITCH;
	  };

	  ($event->type eq 'end_subdoc') && do {
	      my $entity = $event->data;     # An object of class SGMLS_Entity
	      #[[your code for the end of a subdoc entity]]
	      last SWITCH;
	  };

	  ($event->type eq 'conforming') && do {
	      $self->{conforming} = 1;
	      last SWITCH;
	  };

	  # all other events are content
	  push (@{$instance->[4]}, $event_info);
      }
    }

    if ($#{$instance->[4]} < 0) {
	warn "$::prog: empty input from \`$Text::GDO::NSGMLS $file_name'\n";
	return 0;
    } else {
	return 1;
    }
}

#
# Attempt to identify the SGML file named in `file_name'
#
sub identify {
    my ($self, $file_name) = @_;
    my ($sgml_header) = "";

    open (TYPETEST, $file_name)
	|| die "$::prog: opening $file_name: $!\n";
    read (TYPETEST, $sgml_header, 4096)
	|| die "$::prog: reading from $file_name: $!\n";
    close (TYPETEST)
	|| die "$::prog: closing $file_name: $!\n";

    return $self->identify_scalar ($sgml_header);
}

#
# Attempt to identify the SGML file contained in `sgmlHeader'
#
sub identify_scalar {
    my ($self, $sgml_header) = @_;

    # XXX these need to be passed in
    my ($doc_spec);
    my (@dtd_inc) = @INC;

    #
    # Find the type table
    #
    # XXX type tables should be read in elsewhere and cached
    my ($prefix);
    my ($found) = 0;
    my (@keys, %keys);
    foreach $prefix ($self->{'file_dir'}, @dtd_inc) {
	my ($real_filename) = "$prefix/typetab";
	if (-f $real_filename) {
	    my (%types);
	    # Load the type mapping table into `%types', with a
	    # crossreference from `%keys', and listed in order in `@keys'

	    $found = 1;
	    my ($type_tab) = $real_filename;
	    open (TYPETAB, $type_tab)
		|| die "$::prog: opening $type_tab: $!\n";
	    while (<TYPETAB>) {
		next if (/^\s*(#|$)/);
	        chop;
		my ($nick, $type, $default, $key) = split (/\s+/, $_, 4);
	        $nick =~ tr/A-Z/a-z/;
	        if (!defined ($types{$nick})) {
		    $types{$nick} = {nick=>$nick, type=>$type,
				     default=>$default, key=>$key};
		    if (defined $key && $key ne "") {
			$key =~ tr/A-Z/a-z/;
			push (@keys, $key);
			$keys{$key} = $nick;
		    }
		}
	    }
	    close (TYPETAB)
	       || die "$::prog: closing $type_tab: $!\n";
        }
    }
    if (!$found) {
	die "$::prog: Can't find a 'typetab' file in any of the following places:\n    " .
	    join ("\n    ", @dtd_inc) . "\n";
    }

    #
    # If the `doc_spec' is not specified, search the input file for keywords
    #
    if (!defined ($doc_spec)) {
	my ($key);
	my ($ii) = 0;
	my ($rest) = $sgml_header;
      OUTER:
	while (($ii++ < 100) && ($rest =~ m/([^\n]*)\n(.*)/s)) {
	    ($_, $rest) = ($1, $2);
	    foreach $key (@keys) {
		if (/$key/i) {
		    $doc_spec = $keys{$key};
		    last OUTER;
		}
	    }
	}
	die "$::prog: $self->{'file_name'}: unknown document type\n"
	    if (!defined ($doc_spec));

	$self->{'doc_spec'} = $doc_spec;
    }
}

sub doc_spec { return $_[0]->{'doc_spec'} };
sub instance_name { return $_[0]->{'instance_name'} };

package Text::GDO::SGML::Event;
@Text::GDO::SGML::Event::ISA = qw{SGMLS_Event};

sub new {
    my ($type, $sgmls_event) = @_;
    my ($self) = [$sgmls_event->type, $sgmls_event->data, undef, $sgmls_event->line];

    bless $self, $type;
}

sub new_arr {
    my ($type, @arr) = @_;
    my ($self) = \@arr;

    bless $self, $type;
}

package Text::GDO::SGML::Element;
@Text::GDO::SGML::Element::ISA = qw{SGMLS_Element};

sub new {
    my ($type, $name, $parent, @attributes) = @_;

    return bless [$name, $parent, [@attributes]], $type;
}

sub notation {
    return "XXX NOTATION in SGML.pm";
}

sub entity {
    return "XXX ENTITY in SGML.pm";
}

package Text::GDO::SGML::Notation;

sub new {
    my ($type, $name, $sysid, $pubid) = @_;

    return bless [$name, $sysid, $pubid];
}

sub name { return $_[0]->[0]; }
sub sysid { return $_[0]->[1]; }
sub pubid { return $_[0]->[2]; }

package Text::GDO::SGML::Entity;

sub new {
    my ($type, $name, $type, $value, $sysid, $pubid, $filenames, $notation) = @_;
    return bless [$name, $type, $value, {}, $sysid, $pubid, $filenames, $notation];
}

sub name { return $_[0]->[0]; }
sub type { return $_[0]->[1]; }
sub value { return $_[0]->[2]; }
sub data_attributes { return $_[0]->[3]; }
sub sysid { return $_[0]->[4]; }
sub pubid { return $_[0]->[5]; }
sub filenames { return $_[0]->[6]; }
sub notation { return $_[0]->[7]; }

1;
